@@ -10,7 +10,7 @@ package Debian::Debhelper::Buildsystem::pybuild;
 use strict;
 use Dpkg::Control;
 use Dpkg::Changelog::Debian;
-use Debian::Debhelper::Dh_Lib qw(error doit);
+use Debian::Debhelper::Dh_Lib qw(%dh error doit);
 use base 'Debian::Debhelper::Buildsystem';
 
 sub DESCRIPTION {
@@ -28,13 +28,25 @@ sub new {
 	$this->enforce_in_source_building();
 
 	if (!$ENV{'PYBUILD_INTERPRETERS'}) {
-		$this->{pydef} = `pyversions -vd 2>/dev/null`;
+		if ($ENV{'DEBPYTHON_DEFAULT'}) {
+			$this->{pydef} = $ENV{'DEBPYTHON_DEFAULT'};}
+		else {
+			$this->{pydef} = `pyversions -vd 2>/dev/null`;}
 		$this->{pydef} =~ s/\s+$//;
-		$this->{pyvers} = `pyversions -vr 2>/dev/null`;
+		if ($ENV{'DEBPYTHON_SUPPORTED'}) {
+			$this->{pyvers} = $ENV{'DEBPYTHON_SUPPORTED'};}
+		else {
+			$this->{pyvers} = `pyversions -vr 2>/dev/null`;}
 		$this->{pyvers} =~ s/\s+$//;
-		$this->{py3def} = `py3versions -vd 2>/dev/null`;
+		if ($ENV{'DEBPYTHON3_DEFAULT'}) {
+			$this->{py3def} = $ENV{'DEBPYTHON3_DEFAULT'};}
+		else {
+			$this->{py3def} = `py3versions -vd 2>/dev/null`;}
 		$this->{py3def} =~ s/\s+$//;
-		$this->{py3vers} = `py3versions -vr 2>/dev/null`;
+		if ($ENV{'DEBPYTHON3_SUPPORTED'}) {
+			$this->{py3vers} = $ENV{'DEBPYTHON3_SUPPORTED'};}
+		else {
+			$this->{py3vers} = `py3versions -vr 2>/dev/null`;}
 		$this->{py3vers} =~ s/\s+$//;
 		$this->{pypydef} = `pypy -c 'from sys import pypy_version_info as i; print("%s.%s" % (i.major, i.minor))' 2>/dev/null`;
 		$this->{pypydef} =~ s/\s+$//;
@@ -93,13 +105,18 @@ sub pybuild_commands {
 		push @options, '--dir', $dir;
 	}
 
+	if (not grep {$_ eq '--verbose'} @options and $dh{QUIET}) {
+		push @options, '--quiet';
+	}
+
+	my @deps;
 	if ($ENV{'PYBUILD_INTERPRETERS'}) {
 		push @result, ['pybuild', "--$step", @options];
 	}
 	else {
 		# get interpreter packages from Build-Depends{,-Indep}:
 		# NOTE: possible problems with alternative/versioned dependencies
-		my @deps = $this->python_build_dependencies();
+		@deps = $this->python_build_dependencies();
 
 		# When depends on python{3,}-setuptools-scm, set
 		# SETUPTOOLS_SCM_PRETEND_VERSION to upstream version
@@ -112,27 +129,54 @@ sub pybuild_commands {
 			my $version = @{$changelog}[0]->get_version();
 			$version =~ s/-[^-]+$//;  # revision
 			$version =~ s/^\d+://;    # epoch
+			$version =~ s/~/-/;       # ignore tilde versions
 			$ENV{'SETUPTOOLS_SCM_PRETEND_VERSION'} = $version;
 		}
 
+		# When depends on python{3,}-pbr, set PBR_VERSION to upstream version
+		# Without this, python-pbr tries to detect current
+		# version from pkg metadata or git tag, which fails for debian tags
+		# (debian/<version>) sometimes.
+		if ((grep /(pypy|python[0-9\.]*)-pbr/, @deps) && !$ENV{'PBR_VERSION'}) {
+			my $changelog = Dpkg::Changelog::Debian->new(range => {"count" => 1});
+			$changelog->load("debian/changelog");
+			my $version = @{$changelog}[0]->get_version();
+			$version =~ s/-[^-]+$//;  # revision
+			$version =~ s/^\d+://;    # epoch
+			$ENV{'PBR_VERSION'} = $version;
+		}
+
 		my @py2opts = ('pybuild', "--$step");
 		my @py3opts = ('pybuild', "--$step");
+		my @pypyopts = ('pybuild', "--$step");
 
-		if ($step == 'test' and $ENV{'PYBUILD_TEST_PYTEST'} ne '1' and
+		if ($step eq 'test' and $ENV{'PYBUILD_TEST_PYTEST'} ne '1' and
+		       			$ENV{'PYBUILD_TEST_NOSE2'} ne '1' and
 		       			$ENV{'PYBUILD_TEST_NOSE'} ne '1' and
+		       			$ENV{'PYBUILD_TEST_CUSTOM'} ne '1' and
 		       			$ENV{'PYBUILD_TEST_TOX'} ne '1') {
 			if (grep {$_ eq 'python-tox'} @deps and $ENV{'PYBUILD_TEST_TOX'} ne '0') {
 				push @py2opts, '--test-tox'}
 			elsif (grep {$_ eq 'python-pytest'} @deps and $ENV{'PYBUILD_TEST_PYTEST'} ne '0') {
 				push @py2opts, '--test-pytest'}
+			elsif (grep {$_ eq 'python-nose2'} @deps and $ENV{'PYBUILD_TEST_NOSE2'} ne '0') {
+				push @py2opts, '--test-nose2'}
 			elsif (grep {$_ eq 'python-nose'} @deps and $ENV{'PYBUILD_TEST_NOSE'} ne '0') {
 				push @py2opts, '--test-nose'}
-			if (grep {$_ eq 'python3-tox'} @deps and $ENV{'PYBUILD_TEST_TOX'} ne '0') {
+			if (grep {$_ eq 'tox'} @deps and $ENV{'PYBUILD_TEST_TOX'} ne '0') {
 				push @py3opts, '--test-tox'}
 			elsif (grep {$_ eq 'python3-pytest'} @deps and $ENV{'PYBUILD_TEST_PYTEST'} ne '0') {
 				push @py3opts, '--test-pytest'}
+			elsif (grep {$_ eq 'python3-nose2'} @deps and $ENV{'PYBUILD_TEST_NOSE2'} ne '0') {
+				push @py3opts, '--test-nose2'}
 			elsif (grep {$_ eq 'python3-nose'} @deps and $ENV{'PYBUILD_TEST_NOSE'} ne '0') {
 				push @py3opts, '--test-nose'}
+			if (grep {$_ eq 'pypy-tox'} @deps and $ENV{'PYBUILD_TEST_TOX'} ne '0') {
+				push @pypyopts, '--test-tox'}
+			elsif (grep {$_ eq 'pypy-pytest'} @deps and $ENV{'PYBUILD_TEST_PYTEST'} ne '0') {
+				push @pypyopts, '--test-pytest'}
+			elsif (grep {$_ eq 'pypy-nose'} @deps and $ENV{'PYBUILD_TEST_NOSE'} ne '0') {
+				push @pypyopts, '--test-nose'}
 		}
 
 		my $pyall = 0;
@@ -154,10 +198,11 @@ sub pybuild_commands {
 			}
 		}
 		if ($this->{pydef}) {
-			if (not $pyall and grep {$_ eq 'python' or $_ eq 'python-dev'} @deps) {
+			if (not $pyall and grep {$_ eq 'python' or $_ eq 'python-dev' or
+				       		 $_ eq 'python2.7' or $_ eq 'python2.7-dev'} @deps) {
 				push @result, [@py2opts, '-i', $i, '-p', $this->{pydef}, @options];
 			}
-			if (not $pyalldbg and grep {$_ eq 'python-dbg'} @deps) {
+			if (not $pyalldbg and grep {$_ eq 'python-dbg' or $_ eq 'python2.7-dbg'} @deps) {
 				push @result, [@py2opts, '-i', "$i-dbg", '-p', $this->{pydef}, @options];
 			}
 		}
@@ -185,11 +230,14 @@ sub pybuild_commands {
 
 		# PyPy
 		if ($this->{pypydef} and grep {$_ eq 'pypy'} @deps) {
-			push @result, ['pybuild', "--$step", '-i', 'pypy', '-p', $this->{pypydef}, @options];
+			push @result, [@pypyopts, '-i', 'pypy', '-p', $this->{pypydef}, @options];
 		}
 	}
 	if (!@result) {
-		die('E: Please add apropriate interpreter package to Build-Depends, see pybuild(1) for details')
+		use Data::Dumper;
+		die('E: Please add appropriate interpreter package to Build-Depends, see pybuild(1) for details.' .
+		    'this: ' . Dumper($this) .
+		    'deps: ' . Dumper(\@deps));
 	}
 	return @result;
 }
@@ -202,7 +250,7 @@ sub python_build_dependencies {
 	if ($c->load('debian/control')) {
 		for my $field (grep /^Build-Depends/, keys %{$c}) {
 			my $builddeps = $c->{$field};
-			while ($builddeps =~ /(?:^|[\s,])((pypy|python)[0-9\.]*(-[^\s,]+)?)(?:[\s,]|$)/g) {
+			while ($builddeps =~ /(?:^|[\s,])((pypy|python|tox)[0-9\.]*(-[^\s,\(]+)?)(?:[\s,\(]|$)/g) {
 				my $dep = $1;
 				$dep =~ s/:any$//;
 				if ($dep) {push @result, $dep};
